# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for acquisition function maximizers."""

from functools import partial

from boax.optimization.maximizers import functions
from boax.optimization.maximizers.base import Maximizer
from boax.typing import Array, Numeric


def bfgs(
  bounds: Array,
  q: int,
  num_restarts: int,
  num_raw_samples: int,
  eta: Numeric = 1.0,
) -> Maximizer:
  """
  The BFGS acquisition function maximizer.

  Example:
    >>> acqf = upper_confidence_bound(2.0, surrogate)
    >>> maximizer = bfgs(bounds, q=1, num_restarts=25, num_raw_samples=500)
    >>> init_candidates = maximizer.init(key, acqf)
    >>> candidates, values = maximizer.maximize(init_candidates, acqf)

  Args:
    bounds: The bounds of the search space.
    q: The q batch size
    num_restarts: The number of maximization restarts.
    num_initial_samples: The number of initial samples used by the maximizer.
    eta: The temperature parameter.

  Returns:
    The corresponding `Maximizer`.
  """

  return Maximizer(
    init=partial(
      functions.initialization.q_batch_initialization,
      bounds=bounds,
      q=q,
      num_restarts=num_restarts,
      num_raw_samples=num_raw_samples,
      eta=eta,
    ),
    maximize=partial(
      functions.scipy.bfgs,
      bounds=bounds,
    ),
  )



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gaussian Process functions."""

from typing import Tuple

from jax import numpy as jnp
from jax import scipy

from boax.prediction.kernels.base import Kernel
from boax.prediction.means.base import Mean
from boax.typing import Array, Numeric


def prior(
  index_points: Array,
  mean: Mean,
  kernel: Kernel,
  noise: Numeric,
  jitter: Numeric,
) -> Tuple[Array, Array]:
  Kxx = kernel(index_points, index_points)
  loc = mean(index_points)
  cov = Kxx + (noise + jitter) * jnp.identity(Kxx.shape[-1])
  return loc, cov


def posterior(
  index_points: Array,
  observation_index_points: Array,
  observations: Array,
  mean: Mean,
  kernel: Kernel,
  noise: Numeric,
  jitter: Numeric,
) -> Tuple[Array, Array]:
  mz = mean(index_points)
  mx = mean(observation_index_points)

  Kxx = kernel(observation_index_points, observation_index_points)
  Kxz = kernel(observation_index_points, index_points)
  Kzz = kernel(index_points, index_points)

  K = Kxx + (noise + jitter) * jnp.identity(observation_index_points.shape[0])
  chol = scipy.linalg.cholesky(K, lower=True)
  kinvy = scipy.linalg.solve_triangular(
    chol.T, scipy.linalg.solve_triangular(chol, observations - mx, lower=True)
  )
  v = scipy.linalg.solve_triangular(chol, Kxz, lower=True)

  loc = mz + jnp.dot(Kxz.T, kinvy)
  cov = Kzz - jnp.dot(v.T, v) + jitter * jnp.identity(index_points.shape[0])

  return loc, cov


# def variational(
#   index_points: Array,
#   inducing_points: Array,
#   variational_mean: Array,
#   variational_root_cov: Array,
#   mean: Mean,
#   kernel: Kernel,
#   jitter: Numeric,
# ) -> Tuple[Array, Array]:
#   mx = mean(inducing_points)
#   mz = mean(index_points)

#   Kxx = kernel(inducing_points, inducing_points)
#   Kxz = kernel(inducing_points, index_points)
#   Kzz = kernel(index_points, index_points)

#   Lz = jnp.linalg.cholesky(Kxx + jitter * jnp.identity(inducing_points.shape[0]), lower=True)
#   Lz_inv_Kxz = scipy.linalg.solve_triangular(Lz, Kxz, lower=True)
#   Kxx_inv_Kxz = scipy.linalg.solve_triangular(Lz.T, Lz_inv_Kxz)
#   Kxx_inv_Kxz_sqrt = jnp.matmul(Kxx_inv_Kxz.T, variational_root_cov)

#   loc = mz + jnp.matmul(Kxx_inv_Kxz.T, variational_mean - mx)
#   cov = (
#     Kzz
#     - jnp.matmul(Lz_inv_Kxz.T, Lz_inv_Kxz)
#     + jnp.matmul(Kxx_inv_Kxz_sqrt, Kxx_inv_Kxz_sqrt.T)
#   ) + jitter * jnp.identity(inducing_points.shape[0])

#   return loc, cov



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interfaces for kernels."""

from typing import Callable, Protocol

from jax import jit, vmap

from boax.typing import Array, Numeric


class Kernel(Protocol):
  """
  A callable type for kernels.

  A kernel takes two `n x d`-dim arrays as input and returns
  an `n x n`-dim matrix of kernel function values.
  """

  def __call__(self, x: Array, y: Array) -> Array:
    """
    Calculates the kernel function of a pair of inputs.

    Args:
      x: The first `n x d`-dim array.
      y: The second `n x d`-dim array.

    Returns:
      A matrix indicating the kernel function values.
    """


def from_kernel_function(
  kernel_fn: Callable[[Numeric, Numeric], Numeric],
) -> Kernel:
  """
  Transforms a kernel function into a _kernel_.

  Args:
    kernel_fn: The kernel function to be transformed.

  Returns:
    The corresponding `Kernel`.
  """

  return jit(vmap(vmap(kernel_fn, in_axes=(None, 0)), in_axes=(0, None)))



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for acquisition functions."""

import math
from functools import partial
from operator import itemgetter
from typing import Tuple

from jax import jit, lax, scipy

from boax.optimization.acquisitions import functions
from boax.optimization.acquisitions.base import Acquisition
from boax.prediction.models.base import Model
from boax.typing import Array, Numeric
from boax.util import compose, tupled


def probability_of_improvement(
  best: Numeric, model: Model[Tuple[Array, Array]]
) -> Acquisition:
  """
  The Probability of Improvement acquisition function.

  Probability of improvement over the best function value observed so far.

  `PI(x) = P(y >= best_f), y ~ f(x)`

  Example:
    >>> acqf = probability_of_improvement(0.2, surrogate)
    >>> poi = acqf(xs)

  Args:
    best: The best function value observed so far.
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """

  return jit(
    compose(
      scipy.stats.norm.cdf,
      tupled(partial(functions.analytic.scaled_improvement, best=best)),
      tupled(functions.analytic.analytic),
      model,
    )
  )


def log_probability_of_improvement(
  best: Numeric, model: Model[Tuple[Array, Array]]
) -> Acquisition:
  """
  The Log Probability of Improvement acquisition function.

  Logarithm of the probability of improvement over the best function value observed so far.

  `logPI(x) = log(P(y >= best_f)), y ~ f(x)`

  Example:
    >>> acqf = log_probability_of_improvement(0.2, surrogate)
    >>> log_poi = acqf(xs)

  Args:
    best: The best function value observed so far.
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """

  return jit(
    compose(
      scipy.stats.norm.logcdf,
      tupled(partial(functions.analytic.scaled_improvement, best=best)),
      tupled(functions.analytic.analytic),
      model,
    )
  )


def expected_improvement(
  best: Numeric, model: Model[Tuple[Array, Array]]
) -> Acquisition:
  """
  The Expected Improvement acquisition function.

  Expected improvement over the best function value observed so far.

  `EI(x) = E(max(f(x) - best_f, 0))`,

  where the expectation is taken over the value of stochastic function `f` at `x`.

  Example:
    >>> acqf = expected_improvement(0.2, surrogate)
    >>> ei = acqf(xs)

  Args:
    best: The best function value observed so far.
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """

  return jit(
    compose(
      tupled(partial(functions.analytic.ei, best=best)),
      tupled(functions.analytic.analytic),
      model,
    )
  )


def log_expected_improvement(
  best: Numeric, model: Model[Tuple[Array, Array]]
) -> Acquisition:
  """
  The Log Expected Improvement acquisition function.

  Logarithm of the expected improvement over the best function value observed so far.

  `LogEI(x) = log(E(max(f(x) - best_f, 0)))`,

  where the expectation is taken over the value of stochastic function `f` at `x`.

  References:
    Ament, Sebastian, et al. "Unexpected improvements to expected improvement for bayesian optimization."
    arXiv preprint arXiv:2310.20708 (2023).

  Example:
    >>> acqf = log_expected_improvement(0.2, surrogate)
    >>> log_ei = acqf(xs)

  Args:
    best: The best function value observed so far.
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """

  return jit(
    compose(
      tupled(partial(functions.analytic.lei, best=best)),
      tupled(functions.analytic.analytic),
      model,
    )
  )


def upper_confidence_bound(
  beta: Numeric, model: Model[Tuple[Array, Array]]
) -> Acquisition:
  """
  The Upper Confidence Bound (UCB) acquisition function.

  Upper confidence bound comprises of the posterior mean plus an additional term:
  the posterior standard deviation weighted by a trade-off parameter, `beta`.

  `UCB(x) = loc(x) + sqrt(beta) * scale(x)`

  Example:
    >>> acqf = upper_confidence_bound(2.0, surrogate)
    >>> ucb = acqf(xs)

  Args:
    beta: The mean and covariance trade-off parameter.
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """

  return jit(
    compose(
      tupled(partial(functions.analytic.ucb, beta=beta)),
      tupled(functions.analytic.analytic),
      model,
    )
  )


def posterior_mean(
  model: Model[Tuple[Array, Array]],
) -> Acquisition:
  """
  The Posterior mean acquisition function.

  Example:
    >>> acqf = posterior_mean(surrogate)
    >>> mean = acqf(xs)

  Args:
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """

  return jit(
    compose(
      itemgetter(0),
      tupled(functions.analytic.analytic),
      model,
    )
  )


def posterior_scale(
  model: Model[Tuple[Array, Array]],
) -> Acquisition:
  """
  The Posterior scale acquisition function.

  Example:
    >>> acqf = posterior_scale(surrogate)
    >>> scale = acqf(xs)

  Args:
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """

  return jit(
    compose(
      itemgetter(1),
      tupled(functions.analytic.analytic),
      model,
    )
  )


def q_expected_improvement(
  best: Numeric,
  base_samples: Array,
  model: Model[Tuple[Array, Array]],
) -> Acquisition:
  """
  MC-based batch Expected Improvement acquisition function.

  This computes qEI by (1) sampling the joint posterior over q points
  (2) evaluating the improvement over the current best for each sample
  (3) maximizing over q (4) averaging over the samples.

  `qEI(x) = E(max(max y - best_f, 0)), y ~ f(x), x = (x_1,...,x_q)`

  Example:
    >>> acqf = q_expected_improvement(0.2, base_samples, surrogate)
    >>> qei = acqf(xs)

  Args:
    best: The best function value observed so far.
    base_samples: A set of samples from standard normal distribution.
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """

  return jit(
    compose(
      partial(functions.monte_carlo.qei, best=best),
      tupled(partial(functions.monte_carlo.sampler, base_samples=base_samples)),
      model,
    )
  )


def q_probability_of_improvement(
  best: Numeric,
  tau: Numeric,
  base_samples: Array,
  model: Model[Tuple[Array, Array]],
) -> Acquisition:
  """
  MC-based batch Probability of Improvement acquisition function.

  Estimates the probability of improvement over the current best observed value
  by sampling from the joint posterior distribution of the q-batch. MC-based
  estimates of a probability involves taking expectation of an indicator function;
  to support auto-differentiation, the indicator is replaced with a sigmoid
  function with temperature parameter tau.

  `qPI(x) = P(max y >= best_f), y ~ f(x), x = (x_1,...,x_q)`

  Example:
    >>> acqf = q_probability_of_improvement(0.2, 1.0, base_samples, surrogate)
    >>> qpoi = acqf(xs)

  Args:
    best: The best function value observed so far.
    base_samples: A set of samples from standard normal distribution.
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """

  return jit(
    compose(
      partial(functions.monte_carlo.qpoi, best=best, tau=tau),
      tupled(partial(functions.monte_carlo.sampler, base_samples=base_samples)),
      model,
    )
  )


def q_upper_confidence_bound(
  beta: Numeric,
  base_samples: Array,
  model: Model[Tuple[Array, Array]],
) -> Acquisition:
  """
  MC-based batch Upper Confidence Bound acquisition function.

  `qUCB = E(max(mean + |y_tilde - mean|))`,

  where `y_tilde ~ N(mean(x), beta * pi/2 * cov(x))` and `f(x) ~ N(mean(x), cov(x)).`

  Example:
    >>> acqf = q_upper_confidence_bound(2.0, base_samples, surrogate)
    >>> qucb = acqf(xs)

  Args:
    beta: The mean and covariance trade-off parameter.
    base_samples: A set of samples from standard normal distribution.
    model: A gaussian process regression surrogate model.

  Returns:
    The corresponding `Acquisition`.
  """
  beta_prime = lax.sqrt(beta * math.pi / 2)

  return jit(
    compose(
      partial(functions.monte_carlo.qucb, beta=beta_prime),
      tupled(partial(functions.monte_carlo.sampler, base_samples=base_samples)),
      model,
    )
  )



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for surrogate models."""

from functools import partial
from typing import Tuple

from jax import jit

from boax.prediction.kernels.base import Kernel
from boax.prediction.means.base import Mean
from boax.prediction.models import functions
from boax.prediction.models.base import Model
from boax.typing import Array, Numeric


def gaussian_process(
  mean: Mean, kernel: Kernel, noise: Numeric, jitter: Numeric = 1e-6
) -> Model[Tuple[Array, Array]]:
  """
  The gaussian process model.

  Example:
    >>> model = gaussian_process(mean, kernel, 1e-4)
    >>> mean, cov = model(xs)

  Args:
    mean: The process' mean function.
    kernel: The process' covariance function.
    noise: The noise variance in the Normal likelihood distribution of the model.
    jitter: The scalar added to the diagonal of the covariance matrix to ensure positive definiteness.

  Returns:
    The gaussian process `Model`.
  """

  return jit(
    partial(
      functions.gaussian.prior,
      mean=mean,
      kernel=kernel,
      noise=noise,
      jitter=jitter,
    )
  )


def gaussian_process_regression(
  observation_index_points: Array,
  observations: Array,
  mean: Mean,
  kernel: Kernel,
  noise: Numeric,
  jitter: Numeric = 1e-6,
) -> Model[Tuple[Array, Array]]:
  """
  The gaussian process regression model.

  Example:
    >>> model = gaussian_process_regression(x_train, y_train, mean, kernel, 1e-4)
    >>> mean, cov = model(xs)

  Args:
    observation_index_points: The index observation points.
    observations: The values at the index observation points.
    mean: The process' mean function.
    kernel: The process' covariance function.
    noise: The noise variance in the Normal likelihood distribution of the model.
    jitter: The scalar added to the diagonal of the covariance matrix to ensure positive definiteness.

  Returns:
    The gaussian process regression `Model`.
  """

  return jit(
    partial(
      functions.gaussian.posterior,
      observation_index_points=observation_index_points,
      observations=observations,
      mean=mean,
      kernel=kernel,
      noise=noise,
      jitter=jitter,
    )
  )



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for acquisition function maximizers."""

from typing import NamedTuple, Protocol, Tuple

from boax.optimization.acquisitions.base import Acquisition
from boax.typing import Array, PRNGKey


class MaximizerInitFn(Protocol):
  """
  A callable type for the `init` step of a `Maximizer`.

  The `init` step takes PRNG key and an `Acquisition` function
  to construct the initial `n x q x d`-dim candidates for the maximizer.
  """

  def __call__(self, key: PRNGKey, acquisition: Acquisition) -> Array:
    """
    The `init` function.

    Args:
      key: The pseudo-random number generator key.
      acquisition: The acquisition function.

    Returns:
      An initial set of `n x q x d`-dim candidates.
    """


class MaximizerMaximizationFn(Protocol):
  """
  A callable type for the `maximize` step of a `Maximizer`.

  The `maximizer` step takes a set of `n x q x d`-dim initial candidates
  and an `Acquisition` function to find the function's maxima.
  """

  def __call__(
    self, candidates: Array, acquisition: Acquisition
  ) -> Tuple[Array, Array]:
    """
    The `maximize` function.

    Args:
      candidates: The set of initial candidates.
      acquisition: The acquisition function.

    Returns:
      A tuple of maxima candidates and their acquisition values.
    """


class Maximizer(NamedTuple):
  """
  A pair of pure functions implementing acqusition function maximization.

  Boax acquisition function maximizers are all implemented as _maximizer_.
  A maximizer is defined as a pair of pure functions, which are combined
  together in  a `NamedTuple` so that they can be referred to by name.

  Attributes:
    init: A pure function which, when called with a PRNG key and an
      `Acquisition` function, returns a set of initial candidates
      for the acqusition function maximization process.
    maximize: A pure function which takes as input an initial set of
      maxima candidates and an `Acquisition` function to find. The
      maximize function then returns the found maxima and their
      corresponding acquisition values.
  """

  init: MaximizerInitFn
  maximize: MaximizerMaximizationFn



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for search space sampling functions."""

import numpy as np


def primes_less_than(n) -> np.ndarray:
  """
  Sorted array of primes such that `2 <= prime < n`.

  Args:
    n: The upper bound for primes.

  Returns:
    The sorted array of primes.
  """

  j = 3
  primes = np.ones((n + 1) // 2, dtype=bool)

  while j * j <= n:
    if primes[j // 2]:
      primes[j * j // 2 :: j] = False
    j += 2

  ret = 2 * np.where(primes)[0] + 1
  ret[0] = 2  # :(

  return ret



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for samplers."""

from typing import Protocol

from boax.typing import Array, PRNGKey


class Sampler(Protocol):
  """
  A callable type for sampling functions.

  A sampler takes a PRNG key and a number of results as input
  and returns `num_results` samples.
  """

  def __call__(self, key: PRNGKey, num_results: int) -> Array:
    """
    Draws `num_results` of samples.

    Args:
      key: The pseudo-random number generator key.
      candidates: The number of results to return.

    Returns:
      A set of `num_results` samples.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for acquisition functions."""

from typing import Protocol

from boax.typing import Array


class Acquisition(Protocol):
  """
  A callable type for acquisition functions.

  An acquisition function takes a `n x q x d`-dim candidate set as input
  and returns an `n`-dim array of acquisition values.
  """

  def __call__(self, candidates: Array) -> Array:
    """
    Evaluates the acquisition function on a set of `candidates`.

    Args:
      candidates: The `n x q x d`-dim candidate set.

    Returns:
      The `n`-dim acquisition function values at the given design points.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for sampling functions."""

from functools import partial

from jax import lax
from jax import numpy as jnp

from boax.optimization.samplers import functions
from boax.optimization.samplers.base import Sampler
from boax.typing import Array
from boax.util import compose


def halton_uniform(
  minval: Array = jnp.zeros((1,)), maxval: Array = jnp.ones((1,))
) -> Sampler:
  """
  The quasi-MC uniform sampler based on halton sequences.

  Example:
    >>> sampler = halton_uniform(jnp.zeros((5,)), jnp.ones((5,)))
    >>> base_samples = sampler(key, 128)

  Args:
    minval: The inclusive minimum value.
    maxval: The exclusive maximum value.

  Returns:
    The corresponding `Sampler`.
  """

  out_shape = lax.broadcast_shapes(minval.shape, maxval.shape)

  if out_shape[0] < 1 or out_shape[0] > functions.quasi_random.MAX_DIMENSION:
    raise ValueError(
      'Dimension must be between 1 and {}. Supplied {}'.format(
        functions.quasi_random.MAX_DIMENSION, out_shape[0]
      )
    )

  return compose(
    partial(functions.quasi_random.uniform, minval=minval, maxval=maxval),
    partial(functions.quasi_random.halton_sequence, ndims=out_shape[0]),
  )


def halton_normal(
  loc: Array = jnp.zeros((1,)), scale: Array = jnp.ones((1,))
) -> Sampler:
  """
  The quasi-MC normal sampler based on halton sequences.

  Example:
    >>> sampler = halton_normal(jnp.zeros((5,)), jnp.ones((5,)))
    >>> base_samples = sampler(key, 128)

  Args:
    loc: The location parameter.
    scale: The scale parameter.

  Returns:
    The corresponding `Sampler`.
  """

  out_shape = lax.broadcast_shapes(loc.shape, scale.shape)

  if out_shape[0] < 1 or out_shape[0] > functions.quasi_random.MAX_DIMENSION:
    raise ValueError(
      'Dimension must be between 1 and {}. Supplied {}'.format(
        functions.quasi_random.MAX_DIMENSION, out_shape[0]
      )
    )

  return compose(
    partial(functions.quasi_random.normal, loc=loc, scale=scale),
    partial(functions.quasi_random.halton_sequence, ndims=out_shape[0]),
  )



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The kernels sub-package."""

from .alias import matern_five_halves as matern_five_halves
from .alias import matern_one_half as matern_one_half
from .alias import matern_three_halves as matern_three_halves
from .alias import periodic as periodic
from .alias import rbf as rbf
from .base import Kernel as Kernel
from .base import from_kernel_function as from_kernel_function
from .transformed import combine as combine
from .transformed import scale as scale



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for kernels."""

from jax import numpy as jnp

from boax.typing import Array


def squared_distance(x: Array, y: Array) -> Array:
  """
  Computes the squared distance between vectors `x` and `y`.

  Args:
    x: A vector.
    y: A vector.

  Returns:
    Squared distance.
  """

  return jnp.sum((x - y) ** 2)


def euclidean_distance(x: Array, y: Array) -> Array:
  """
  Computes the euclidean distance between vectors `x` and `y`.

  Args:
    x: A vector.
    y: A vector.

  Returns:
    Euclidean distance.
  """

  pdist = x - y
  is_zero = jnp.allclose(pdist, 0.0)
  masked_pdist = jnp.where(is_zero, jnp.ones_like(pdist), pdist)
  euclidean = jnp.linalg.norm(masked_pdist, ord=2)

  return jnp.where(is_zero, 0, euclidean)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The maximizer initialization functions."""

from jax import nn, random
from jax import numpy as jnp

from boax.optimization.acquisitions.base import Acquisition
from boax.optimization.samplers.alias import halton_uniform
from boax.typing import Array, Numeric, PRNGKey


def q_batch_initialization(
  key: PRNGKey,
  acquisition: Acquisition,
  bounds: Array,
  q: int,
  num_restarts: int,
  num_raw_samples: int,
  eta: Numeric,
) -> Array:
  raw_key, sample_key = random.split(key)
  raw_samples = halton_uniform(bounds[:, 0], bounds[:, 1])(
    raw_key, num_raw_samples * q
  )

  candidates = jnp.reshape(raw_samples, (num_raw_samples, q, -1))
  values = acquisition(candidates)
  weights = jnp.exp(eta * nn.standardize(values, axis=0))

  return random.choice(sample_key, candidates, (num_restarts,), p=weights)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Periodic kernels."""

from jax import numpy as jnp

from boax.typing import Numeric


def periodic(
  x: Numeric,
  y: Numeric,
  length_scale: Numeric,
  variance: Numeric,
  period: Numeric,
) -> Numeric:
  sine_squared = (jnp.sin(jnp.pi * (x - y) / period) / length_scale) ** 2
  K = variance * jnp.exp(-0.5 * jnp.sum(sine_squared, axis=0))
  return K



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""General util functions."""

from functools import reduce
from typing import Callable, TypeVar

T = TypeVar('T')


def identity(i: T) -> T:
  """
  Identity Function

  Args:
    i: Input value.

  Returns:
    The input value.
  """

  return i


def const(c: T) -> Callable:
  """
  Constant Function

  Args:
    c: Constant value.

  Returns:
    A function that returns the given constant.
  """

  def __fn(*args, **kwargs) -> T:
    return c

  return __fn


def compose(*fns: Callable) -> Callable:
  """
  Composes a sequence of function

  Args:
    fns: Functions to compose.

  Returns:
    A function composed of the sequential applications of the given functions.
  """

  def __reduce_fn(f: Callable, g: Callable) -> Callable:
    def __fn(*args, **kwargs):
      return f(g(*args, **kwargs))

    return __fn

  return reduce(__reduce_fn, fns)


def tupled(f: Callable) -> Callable:
  """
  Transforms function to take a tuple as its argument.

  Args:
    fns: Functions to transform.

  Returns:
    The transformed function.
  """

  def __fn(x):
    return f(*x)

  return __fn


def untupled(f: Callable) -> Callable:
  """
  Transforms function to take a regular list arguments.

  Args:
    fns: Functions to transform.

  Returns:
    The transformed function.
  """

  def __fn(*xs):
    return f(tuple(xs))

  return __fn



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transformation functions for kernels."""

from typing import Callable

from jax import numpy as jnp

from boax.prediction.kernels.base import Kernel
from boax.typing import Numeric


def scale(amplitude: Numeric, inner: Kernel) -> Kernel:
  """
  Scale kernel.

  Computes `k(x, y) = amplitude * inner(x, y)`.

  Example:
    >>> inner = rbf(jnp.array([0.2, 3.0]))
    >>> kernel = scale(3.0, inner)
    >>> Kxx = kernel(xs, xs)

  Args:
    amplitude: The parameter controlling the maximum of the kernel.
    inner: The inner kernel.

  Returns:
    A scaled `Kernel`.
  """

  def kernel(x, y):
    return amplitude * inner(x, y)

  return kernel


def combine(operator: Callable, *kernels: Kernel) -> Kernel:
  """
  Combine kernel.

  Combines a sequence of kernels using the an `operator`.

  Example:
    >>> inners = list(map(rbf, [0.2, 0.3, 0.4]))
    >>> kernel = combine(partial(jnp.sum, axis=0), *inners)
    >>> Kxx = kernel(xs, xs)

  Args:
    operator: The operator used for the combination.
    kernels: The sequence of inner kernels.

  Returns:
    A combined `Kernel`.
  """

  def kernel(x, y):
    return operator(jnp.stack([k(x, y) for k in kernels]))

  return kernel



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for mean functions."""

from jax import jit, vmap
from jax import numpy as jnp

from boax.prediction.means.base import Mean
from boax.typing import Array, Numeric
from boax.util import const


def zero() -> Mean:
  """
  The zero mean function.

  Computes `y = f(x) = 0`.

  Example:
    >>> mean = zero()
    >>> ys = mean(xs)

  Returns:
    A zero mean function.
  """

  return jit(vmap(const(jnp.zeros(()))))


def constant(x: Numeric) -> Mean:
  """
  The constant mean function.

  Computes `y = f(x) = c`.

  Example:
    >>> mean = constant(1.96)
    >>> ys = mean(xs)

  Returns:
    A constant mean function.
  """

  return jit(vmap(const(x)))


def linear(scale: Array, bias: Numeric) -> Mean:
  """
  The linear mean function.

  Computes `y = f(x) = scale * x + bias`.

  Example:
    >>> mean = linear(jnp.array([0.2, 3.0]), 5.0)
    >>> ys = mean(xs)

  Returns:
    A linear mean function.
  """

  def mean(value: Array) -> Array:
    return jnp.dot(scale, value) + bias

  return jit(mean)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The maximizer initialization functions."""

from functools import partial
from typing import Tuple

from jax import numpy as jnp
from jax.scipy import optimize

from boax.optimization.acquisitions.base import Acquisition
from boax.typing import Array
from boax.util import compose


def bfgs(
  candidates: Array, acquisition: Acquisition, bounds: Array
) -> Tuple[Array, Array]:
  results = optimize.minimize(
    fun=compose(
      jnp.negative,
      jnp.sum,
      acquisition,
      partial(jnp.reshape, newshape=candidates.shape),
    ),
    x0=candidates.flatten(),
    method='bfgs',
  )

  clipped = jnp.clip(
    jnp.reshape(results.x, candidates.shape),
    a_min=bounds[:, 0],
    a_max=bounds[:, 1],
  )

  return clipped, acquisition(clipped)
